// Copyright 1992-2010 by Jon Dart.e  All Rights Reserved.

#ifndef _SCORING_H
#define _SCORING_H

#include "board.h"

#include <iostream>
using namespace std;

    struct PawnHashEntry {

       hash_t hc;

       enum {OUTSIDE_PP = 1};

       struct PawnData {
           Bitboard passers;
           byte weak_files;
	   byte weakopen;
           byte pawn_file_mask;
           byte passer_file_mask;
           Bitboard opponent_pawn_attacks;
	   int passer_score;
  	   int center_score;
	   int score;
  	   int w_square_pawns,b_square_pawns;
           int doubled_score, isolated_score;
           int outside;

	   int weakFile(int file) const {
	     return weak_files & (1<<(file-1));
	   }
       } wPawnData, bPawnData;

       const PawnData &pawnData(ColorType side) const {
	 return (side==White) ? wPawnData : bPawnData;
       }

    };

class Scoring
{
    // This class does static evaluation of chess positions.

    public:
		
    enum { INVALID_SCORE = -Constants::MATE-1 };

    static void init();

    static void cleanup();

    Scoring();

    ~Scoring();
        
    // evaluate "board" from the perspective of the side to move.
    int evalu8( const Board &board );

    // evaluate "board" from the perspective of the side to move.
    int evalu8( const Board &board, int alpha, int beta );

    // return a pawn structure entry
    PawnHashEntry *pawnEntry(const Board &board);

    // return a positional score
    int positionalScore( const Board &board, 
			int alpha, int beta);

    // return a material score
    int materialScore( const Board &board );
		
    // checks for legal draws plus certain other theoretically
    // draw positions
    static int isDraw(const Board &board);
        
    // checks for draw by repetition (returning repetition count) + 
    // other draw situations as in isDraw above.
    static int isDraw( const Board &board, int &rep_count, int ply);

    static int repetitionDraw( const Board &board );

    static int materialDraw( const Board &board );

    // Check only for legal draws
    static int isLegalDraw(const Board &board);

    static int theoreticalDraw(const Board &board);

    // Turn a score into a formatted string (mate scores are
    // shown like +Mate6).
    static void printScore( int score, ostream & );

    // Output scores in the format required by the UCI protocol.
    static void printScoreUCI( int score, ostream & );

    static bool mateScore(int score) {
      return score>=Constants::MATE-255 || score<=-Constants::MATE-255;
    }

#ifdef EVAL_STATS
    static void clearStats();
#endif

    // Try to return a score based on bitbases, INVALID_SCORE if not found
    static int tryBitbase(const Board &board);

    void cacheBestMove(const Board &board, Move best);

    Move getBestMove(const Board &board) const;

    void clearHashTables();

 private:

    PawnHashEntry *pawnHashTable;

    struct KingCoverHashEntry {
      hash_t hc;
      int cover;
      int file_exposure;
    };

    KingCoverHashEntry *kingCoverHashTable;

    struct EndgameHashEntry {
        // defines one entry in the king/pawn hash table.
        hash_t hc;                                    // hashcode
        signed char white_king_position, black_king_position;
        signed char white_endgame_pawn_proximity;
        signed char black_endgame_pawn_proximity;
        byte w_uncatchable, b_uncatchable;
        byte pawns_both_sides;
        int wScore,bScore;
    };

    EndgameHashEntry *endgameHashTable;

    // The scores for opening, middlegame and endgame
    struct Scores {

      Scores()
        :mid(0), end(0),any(0)
      {
      }
      int mid, end, any; 
    };

    void pieceScore(const Board &board, ColorType side,
                   const PawnHashEntry::PawnData &ourPawnData,
		   const PawnHashEntry::PawnData &oppPawnData,
                    int cover, Scores &, bool endgame);
    int calcPawnData(const Board &board, ColorType side,
			   PawnHashEntry::PawnData &entr);
    static int calcCover(const Board &board, ColorType side, Square kp);
    void calcCover(const Board &board, ColorType side,
		   KingCoverHashEntry &cover);
    void evalOutsidePassers(const Board &board,
			   PawnHashEntry *pawnEntry);
    void calcPawnEntry(const Board &board, PawnHashEntry *pawnEntry);

    void pawnScore(const Board &board, ColorType side,
		  const PawnHashEntry::PawnData &oppPawnData, Scores &);

    void calcEndgame(const Board &board, PawnHashEntry *pawnEntry,
		     EndgameHashEntry *endgameEntry);

    void scoreEndgame(const Board &, EndgameHashEntry *endgameEntry,
		     const PawnHashEntry::PawnData &pawnData, ColorType side,
		     Scores &);

    int outpost(const Board &board, ColorType side, Square sq, Square scoreSq, 
                const int scores[64],
                const PawnHashEntry::PawnData &oppPawnData);

};

#endif



